@@ -1,5 +1,6 @@
 import sys
 from collections import OrderedDict
+from contextlib import asynccontextmanager
 from functools import partial
 from ipaddress import ip_address
 import itertools
@@ -8,10 +9,11 @@ import random
 import ssl
 import struct
 import urllib.parse
+from typing import List, Optional
 
-from async_generator import async_generator, yield_, asynccontextmanager
 import trio
 import trio.abc
+from exceptiongroup import BaseExceptionGroup
 from wsproto import ConnectionType, WSConnection
 from wsproto.connection import ConnectionState
 import wsproto.frame_protocol as wsframeproto
@@ -28,6 +30,7 @@ from wsproto.events import (
 )
 import wsproto.utilities
 
+_TRIO_MULTI_ERROR = tuple(map(int, trio.__version__.split('.'))) < (0, 22, 0)
 
 CONN_TIMEOUT = 60 # default connect & disconnect timeout, in seconds
 MESSAGE_QUEUE_SIZE = 1
@@ -36,6 +39,10 @@ RECEIVE_BYTES = 4 * 2 ** 10 # 4 KiB
 logger = logging.getLogger('trio-websocket')
 
 
+def _ignore_cancel(exc):
+    return None if isinstance(exc, trio.Cancelled) else exc
+
+
 class _preserve_current_exception:
     """A context manager which should surround an ``__exit__`` or
     ``__aexit__`` handler or the contents of a ``finally:``
@@ -58,14 +65,16 @@ class _preserve_current_exception:
         if value is None or not self._armed:
             return False
 
-        def remove_cancels(exc):
-            return None if isinstance(exc, trio.Cancelled) else exc
-
-        return trio.MultiError.filter(remove_cancels, value) is None
+        if _TRIO_MULTI_ERROR:
+            filtered_exception = trio.MultiError.filter(_ignore_cancel, value)  # pylint: disable=no-member
+        elif isinstance(value, BaseExceptionGroup):
+            filtered_exception = value.subgroup(lambda exc: not isinstance(exc, trio.Cancelled))
+        else:
+            filtered_exception = _ignore_cancel(value)
+        return filtered_exception is None
 
 
 @asynccontextmanager
-@async_generator
 async def open_websocket(host, port, resource, *, use_ssl, subprotocols=None,
     extra_headers=None,
     message_queue_size=MESSAGE_QUEUE_SIZE, max_message_size=MAX_MESSAGE_SIZE,
@@ -115,7 +124,7 @@ async def open_websocket(host, port, resource, *, use_ssl, subprotocols=None,
         except OSError as e:
             raise HandshakeError from e
         try:
-            await yield_(connection)
+            yield connection
         finally:
             try:
                 with trio.fail_after(disconnect_timeout):
@@ -175,9 +184,10 @@ async def connect_websocket(nursery, host, port, resource, *, use_ssl,
     if port in (80, 443):
         host_header = host
     else:
-        host_header = '{}:{}'.format(host, port)
-    wsproto = WSConnection(ConnectionType.CLIENT)
-    connection = WebSocketConnection(stream, wsproto, host=host_header,
+        host_header = f'{host}:{port}'
+    connection = WebSocketConnection(stream,
+        WSConnection(ConnectionType.CLIENT),
+        host=host_header,
         path=resource,
         client_subprotocols=subprotocols, client_extra_headers=extra_headers,
         message_queue_size=message_queue_size,
@@ -291,6 +301,11 @@ def _url_to_host(url, ssl_context):
     else:
         port = 443 if ssl_context else 80
     path_qs = parts.path
+    # RFC 7230, Section 5.3.1:
+    # If the target URI's path component is empty, the client MUST
+    # send "/" as the path within the origin-form of request-target.
+    if not path_qs:
+        path_qs = '/'
     if '?' in url:
         path_qs += '?' + parts.query
     return host, port, path_qs, ssl_context
@@ -324,8 +339,9 @@ async def wrap_client_stream(nursery, stream, host, resource, *,
         then the connection is closed with code 1009 (Message Too Big).
     :rtype: WebSocketConnection
     '''
-    wsproto = WSConnection(ConnectionType.CLIENT)
-    connection = WebSocketConnection(stream, wsproto, host=host, path=resource,
+    connection = WebSocketConnection(stream,
+        WSConnection(ConnectionType.CLIENT),
+        host=host, path=resource,
         client_subprotocols=subprotocols, client_extra_headers=extra_headers,
         message_queue_size=message_queue_size,
         max_message_size=max_message_size)
@@ -352,8 +368,8 @@ async def wrap_server_stream(nursery, stream,
     :type stream: trio.abc.Stream
     :rtype: WebSocketRequest
     '''
-    wsproto = WSConnection(ConnectionType.SERVER)
-    connection = WebSocketConnection(stream, wsproto,
+    connection = WebSocketConnection(stream,
+        WSConnection(ConnectionType.SERVER),
         message_queue_size=message_queue_size,
         max_message_size=max_message_size)
     nursery.start_soon(connection._reader_task)
@@ -442,7 +458,7 @@ class ConnectionClosed(Exception):
 
     def __repr__(self):
         ''' Return representation. '''
-        return '{}<{}>'.format(self.__class__.__name__, self.reason)
+        return f'{self.__class__.__name__}<{self.reason}>'
 
 
 class ConnectionRejected(HandshakeError):
@@ -467,8 +483,7 @@ class ConnectionRejected(HandshakeError):
 
     def __repr__(self):
         ''' Return representation. '''
-        return '{}<status_code={}>'.format(self.__class__.__name__,
-            self.status_code)
+        return f'{self.__class__.__name__}<status_code={self.status_code}>'
 
 
 class CloseReason:
@@ -478,7 +493,7 @@ class CloseReason:
         Constructor.
 
         :param int code:
-        :param str reason:
+        :param Optional[str] reason:
         '''
         self._code = code
         try:
@@ -511,8 +526,8 @@ class CloseReason:
 
     def __repr__(self):
         ''' Show close code, name, and reason. '''
-        return '{}<code={}, name={}, reason={}>'.format(self.__class__.__name__,
-            self.code, self.name, self.reason)
+        return f'{self.__class__.__name__}' \
+               f'<code={self.code}, name={self.name}, reason={self.reason}>'
 
 
 class Future:
@@ -615,7 +630,7 @@ class WebSocketRequest:
         :rtype: WebSocketConnection
         '''
         if extra_headers is None:
-            extra_headers = list()
+            extra_headers = []
         await self._connection._accept(self._event, subprotocol, extra_headers)
         return self._connection
 
@@ -647,14 +662,12 @@ def _get_stream_endpoint(stream, *, local):
         represented as an endpoint.
     :rtype: Endpoint or str
     '''
+    socket, is_ssl = None, False
     if isinstance(stream, trio.SocketStream):
         socket = stream.socket
-        is_ssl = False
     elif isinstance(stream, trio.SSLStream):
         socket = stream.transport_stream.socket
         is_ssl = True
-    else:
-        socket = None
     if socket:
         addr, port, *_ = socket.getsockname() if local else socket.getpeername()
         endpoint = Endpoint(addr, port, is_ssl)
@@ -668,7 +681,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
 
     CONNECTION_ID = itertools.count()
 
-    def __init__(self, stream, wsproto, *, host=None, path=None,
+    def __init__(self, stream, ws_connection, *, host=None, path=None,
         client_subprotocols=None, client_extra_headers=None,
         message_queue_size=MESSAGE_QUEUE_SIZE,
         max_message_size=MAX_MESSAGE_SIZE):
@@ -684,7 +697,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
         for you.
 
         :param SocketStream stream:
-        :type wsproto: wsproto.WSConnection
+        :param ws_connection wsproto.WSConnection:
         :param str host: The hostname to send in the HTTP request headers. Only
             used for client connections.
         :param str path: The URL path for this connection.
@@ -702,12 +715,12 @@ class WebSocketConnection(trio.abc.AsyncResource):
         self._id = next(self.__class__.CONNECTION_ID)
         self._stream = stream
         self._stream_lock = trio.StrictFIFOLock()
-        self._wsproto = wsproto
+        self._wsproto = ws_connection
         self._message_size = 0
         self._message_parts = []  # type: List[bytes|str]
         self._max_message_size = max_message_size
         self._reader_running = True
-        if wsproto.client:
+        if ws_connection.client:
             self._initial_request = Request(host=host, target=path,
                 subprotocols=client_subprotocols,
                 extra_headers=client_extra_headers or [])
@@ -731,9 +744,10 @@ class WebSocketConnection(trio.abc.AsyncResource):
         # Set once a WebSocket closed handshake takes place, i.e after a close
         # frame has been sent and a close frame has been received.
         self._close_handshake = trio.Event()
-        # Set immediately upon receiving closed event from peer.  Used to
-        # test close race conditions between client and server.
-        self._for_testing_peer_closed_connection = trio.Event()
+        # When not None, set immediately upon receiving CloseConnection from peer,
+        # followed by sleep(0) for a context switch.  Used to test close race
+        # conditions between client and server.
+        self._for_testing_peer_closed_connection: Optional[trio.Event] = None
 
     @property
     def closed(self):
@@ -824,17 +838,22 @@ class WebSocketConnection(trio.abc.AsyncResource):
         with _preserve_current_exception():
             await self._aclose(code, reason)
 
-    async def _aclose(self, code=1000, reason=None):
+    async def _aclose(self, code, reason):
         if self._close_reason:
             # Per AsyncResource interface, calling aclose() on a closed resource
             # should succeed.
             return
         try:
             if self._wsproto.state == ConnectionState.OPEN:
+                # Our side is initiating the close, so send a close connection
+                # event to peer, while setting the local close reason to normal.
+                self._close_reason = CloseReason(1000, None)
                 await self._send(CloseConnection(code=code, reason=reason))
             elif self._wsproto.state in (ConnectionState.CONNECTING,
                     ConnectionState.REJECTING):
                 self._close_handshake.set()
+            # TODO: shouldn't the receive channel be closed earlier, so that
+            #  get_message() during send of the CloseConneciton event fails?
             await self._recv_channel.aclose()
             await self._close_handshake.wait()
         except ConnectionClosed:
@@ -891,8 +910,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
         if self._close_reason:
             raise ConnectionClosed(self._close_reason)
         if payload in self._pings:
-            raise ValueError('Payload value {} is already in flight.'.
-                format(payload))
+            raise ValueError(f'Payload value {payload} is already in flight.')
         if payload is None:
             payload = struct.pack('!I', random.getrandbits(32))
         event = trio.Event()
@@ -919,7 +937,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
 
         :param message: The message to send.
         :type message: str or bytes
-        :raises ConnectionClosed: if connection is already closed.
+        :raises ConnectionClosed: if connection is closed, or being closed
         '''
         if self._close_reason:
             raise ConnectionClosed(self._close_reason)
@@ -934,7 +952,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
     def __str__(self):
         ''' Connection ID and type. '''
         type_ = 'client' if self.is_client else 'server'
-        return '{}-{}'.format(type_, self._id)
+        return f'{type_}-{self._id}'
 
     async def _accept(self, request, subprotocol, extra_headers):
         '''
@@ -1024,9 +1042,9 @@ class WebSocketConnection(trio.abc.AsyncResource):
         :rtype: WebSocketRequest
         '''
         if not self.is_server:
-            raise Exception('This method is only valid for server connections.')
+            raise RuntimeError('This method is only valid for server connections.')
         if self._connection_proposal is None:
-            raise Exception('No proposal available. Did you call this method'
+            raise RuntimeError('No proposal available. Did you call this method'
                 ' multiple times or at the wrong time?')
         proposal = await self._connection_proposal.wait_value()
         self._connection_proposal = None
@@ -1083,12 +1101,22 @@ class WebSocketConnection(trio.abc.AsyncResource):
 
         :param wsproto.events.CloseConnection event:
         '''
-        self._for_testing_peer_closed_connection.set()
-        await trio.sleep(0)
         if self._wsproto.state == ConnectionState.REMOTE_CLOSING:
+            # Set _close_reason in advance, so that send_message() will raise
+            # ConnectionClosed during the close handshake.
+            self._close_reason = CloseReason(event.code, event.reason or None)
+            if self._for_testing_peer_closed_connection:
+                self._for_testing_peer_closed_connection.set()
+                await trio.sleep(0)
             await self._send(event.response())
         await self._close_web_socket(event.code, event.reason or None)
         self._close_handshake.set()
+        # RFC: "When a server is instructed to Close the WebSocket Connection
+        #   it SHOULD initiate a TCP Close immediately, and when a client is
+        #   instructed to do the same, it SHOULD wait for a TCP Close from the
+        #   server."
+        if self.is_server:
+            await self._close_stream()
 
     async def _handle_message_event(self, event):
         '''
@@ -1100,8 +1128,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
         self._message_size += len(event.data)
         self._message_parts.append(event.data)
         if self._message_size > self._max_message_size:
-            err = 'Exceeded maximum message size: {} bytes'.format(
-                self._max_message_size)
+            err = f'Exceeded maximum message size: {self._max_message_size} bytes'
             self._message_size = 0
             self._message_parts = []
             self._close_reason = CloseReason(1009, err)
@@ -1227,7 +1254,7 @@ class WebSocketConnection(trio.abc.AsyncResource):
 
     async def _send(self, event):
         '''
-        Send an to the remote WebSocket.
+        Send an event to the remote WebSocket.
 
         The reader task and one or more writers might try to send messages at
         the same time, so this method uses an internal lock to serialize
@@ -1266,13 +1293,12 @@ class Endpoint:
         else:
             port_str = ':' + str(self.port)
         if self.address.version == 4:
-            return '{}://{}{}'.format(scheme, self.address, port_str)
-        return '{}://[{}]{}'.format(scheme, self.address, port_str)
+            return f'{scheme}://{self.address}{port_str}'
+        return f'{scheme}://[{self.address}]{port_str}'
 
     def __repr__(self):
         ''' Return endpoint info as string. '''
-        return 'Endpoint(address="{}", port={}, is_ssl={})'.format(self.address,
-            self.port, self.is_ssl)
+        return f'Endpoint(address="{self.address}", port={self.port}, is_ssl={self.is_ssl})'
 
 
 class WebSocketServer:
@@ -1295,8 +1321,8 @@ class WebSocketServer:
         multiple listeners that have _different port numbers!_ See the
         ``listeners`` property.
 
-        :param handler: the async function called with the corresponding
-            WebSocketConnection on each new connection.  The call will be made
+        :param handler: the async function called with a :class:`WebSocketRequest`
+            on each new connection.  The call will be made
             once the HTTP handshake completes, which notably implies that the
             connection's `path` property will be valid.
         :param listeners: The WebSocket will be served on each of the listeners.
@@ -1333,12 +1359,11 @@ class WebSocketServer:
         if len(self._listeners) > 1:
             raise RuntimeError('Cannot get port because this server has'
                 ' more than 1 listeners.')
+        listener = self.listeners[0]
         try:
-            listener = self.listeners[0]
             return listener.port
         except AttributeError:
-            raise RuntimeError('This socket does not have a port: {!r}'
-                .format(listener)) from None
+            raise RuntimeError(f'This socket does not have a port: {repr(listener)}') from None
 
     @property
     def listeners(self):
@@ -1350,16 +1375,14 @@ class WebSocketServer:
         :returns: Listeners
         :rtype list[Endpoint or str]:
         '''
-        listeners = list()
+        listeners = []
         for listener in self._listeners:
+            socket, is_ssl = None, False
             if isinstance(listener, trio.SocketListener):
                 socket = listener.socket
-                is_ssl = False
             elif isinstance(listener, trio.SSLListener):
                 socket = listener.transport_listener.socket
                 is_ssl = True
-            else:
-                socket = None
             if socket:
                 sockname = socket.getsockname()
                 listeners.append(Endpoint(sockname[0], sockname[1], is_ssl))
@@ -1397,8 +1420,8 @@ class WebSocketServer:
         :type stream: trio.abc.Stream
         '''
         async with trio.open_nursery() as nursery:
-            wsproto = WSConnection(ConnectionType.SERVER)
-            connection = WebSocketConnection(stream, wsproto,
+            connection = WebSocketConnection(stream,
+                WSConnection(ConnectionType.SERVER),
                 message_queue_size=self._message_queue_size,
                 max_message_size=self._max_message_size)
             nursery.start_soon(connection._reader_task)
